import { __awaiter } from "tslib";
import { BaseClient } from '@sentry/core';
import { logger, SyncPromise } from '@sentry/utils';
import { SDK_NAME } from '../common';
import { MainBackend } from './backend';
import { addEventDefaults } from './context';
import { normalizeEvent } from './normalize';
/** SDK version used in every event. */
// tslint:disable-next-line
export const SDK_VERSION = require('../../package.json').version;
/** Frontend implementation for Electron renderer backends. */
export class MainClient extends BaseClient {
    /**
     * Creates a new Electron SDK instance.
     * @param options Configuration options for this SDK.
     */
    constructor(options) {
        super(MainBackend, options);
    }
    /**
     * @inheritDoc
     */
    _prepareEvent(event, scope, hint) {
        event.platform = event.platform || 'node';
        event.sdk = Object.assign(Object.assign({}, event.sdk), { name: SDK_NAME, packages: [
                ...((event.sdk && event.sdk.packages) || []),
                {
                    name: 'npm:@sentry/electron',
                    version: SDK_VERSION,
                },
            ], version: SDK_VERSION });
        return super._prepareEvent(event, scope, hint).then((filledEvent) => new SyncPromise((resolve) => __awaiter(this, void 0, void 0, function* () {
            if (filledEvent) {
                resolve(normalizeEvent(yield addEventDefaults(this._options.appName, filledEvent)));
            }
            else {
                resolve(filledEvent);
            }
        })).then((e) => e));
    }
    /**
     * Uploads a native crash dump (Minidump) to Sentry.
     *
     * @param path The relative or absolute path to the minidump.
     * @param event Optional event payload to attach to the minidump.
     * @param scope Optional SDK scope used to upload.
     */
    captureMinidump(path, event = {}, scope) {
        let eventId;
        this._processing = true;
        event.tags = Object.assign({ event_type: 'native' }, event.tags);
        // We are not calling _processEvent here since we do not have beforeSend for minidump crashes
        this._prepareEvent(event, scope, undefined).then(finalEvent => {
            if (finalEvent) {
                eventId = finalEvent && finalEvent.event_id;
                this._getBackend().uploadMinidump(path, finalEvent);
            }
            this._processing = false;
        }, reason => {
            logger.error(reason);
            this._processing = false;
        });
        return eventId;
    }
    /**
     * @inheritDoc
     */
    captureEvent(event, hint, scope) {
        event.tags = Object.assign({ event_type: 'javascript' }, event.tags);
        return super.captureEvent(event, hint, scope);
    }
    /**
     * Does nothing in main/node
     */
    showReportDialog(_) {
        // noop
    }
}
//# sourceMappingURL=client.js.map