#include <napi.h>
#include <stdio.h>
#include <iostream>
#include "filmlab/scene.hpp"
#include "filmlab/sceneview.hpp"
#include "filmlab/rawimage.hpp"
#include "filmlab/image.hpp"
#include "filmlab/imageloader.hpp"
#include "filmlab/imagesaver.hpp"
#include "filmlab/renderer.hpp"
#include "filmlab/gl_helpers.hpp"
#include "filmlab/transforms.hpp"
#include <opencv2/opencv.hpp>
#include <future>
#include <iostream>

Napi::Uint8Array U8ArrayView(Napi::Env env, uint8_t* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements);
    return Napi::Uint8Array::New(env, elements, buffer, 0);
}

Napi::Uint16Array U16ArrayView(Napi::Env env, uint16_t* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements * sizeof(uint16_t));
    return Napi::Uint16Array::New(env, elements, buffer, 0);
}

Napi::Float32Array FloatArrayView(Napi::Env env, float* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements * sizeof(float));
    return Napi::Float32Array::New(env, elements, buffer, 0);
}

Napi::Float64Array DoubleArrayView(Napi::Env env, double* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements * sizeof(double));
    return Napi::Float64Array::New(env, elements, buffer, 0);
}

Napi::Object ImageFormatWrapper(Napi::Env env, filmlab::ImageFormat& formatData) {
  Napi::Object imageFormat = Napi::Object::New(env);
  imageFormat.Set("orientation", static_cast<int>(formatData.orientation));
  imageFormat.Set("colorSpace", static_cast<int>(formatData.colorSpace));
  imageFormat.Set("pixelOrder", static_cast<int>(formatData.pixelOrder));
  imageFormat.Set("gammaCurve", static_cast<int>(formatData.gammaCurve));
  //imageFormat.Set("illuminantXY", DoubleArrayView(env, formatData.illuminantXY, 2));
  //imageFormat.Set("transformMatrixXYZ", DoubleArrayView(env, formatData.transformMatrixXYZ[0], 9));
  return imageFormat;
}

Napi::Object MatWrapper(Napi::Env env, cv::Mat& mat) {
    Napi::Object wrappedMat = Napi::Object::New(env);
    wrappedMat.Set("rows", mat.rows);
    wrappedMat.Set("cols", mat.cols);
    wrappedMat.Set("type", mat.type());
    wrappedMat.Set("channels", mat.channels());
    wrappedMat.Set("depth", mat.depth());
    if (mat.type() == CV_16U || mat.type() == CV_16F) {
      wrappedMat.Set("data", U16ArrayView(env, mat.ptr<uint16>(), mat.total() * mat.channels()));
    } else if (mat.type() == CV_32F) {
      wrappedMat.Set("data", FloatArrayView(env, mat.ptr<float>(), mat.total() * mat.channels()));
    } else if (mat.type() == CV_64F) {
      wrappedMat.Set("data", DoubleArrayView(env, mat.ptr<double>(), mat.total() * mat.channels()));
    } else {
      wrappedMat.Set("data", U8ArrayView(env, mat.ptr<uchar>(), mat.total() * mat.elemSize()));
    }
   
    return wrappedMat;
}

Napi::Object ImageWrapper(Napi::Env env, filmlab::Image& image) {
   Napi::Object wrappedImage = Napi::Object::New(env);
   wrappedImage.Set("format", ImageFormatWrapper(env, image.format));
   wrappedImage.Set("mat", MatWrapper(env, image.mat));
   return wrappedImage;
}

filmlab::Scene* scene = nullptr;
filmlab::SceneView* sceneView;
std::string currentImagePath;

// We used to use ThreadSafeFunction with a callback handler here, but it's
// unreliable on Windows so for now we're using this dumb/simple approach
// instead. See https://github.com/nodejs/node-addon-api/issues/784#issuecomment-670822249
bool rendererDataChanged = false;
bool sceneDataChanged = false;

std::function<void()> sceneChangeHandler = []() {
  sceneDataChanged = true;
};

std::function<void()> rendererChangeHandler = []() {
  rendererDataChanged = true;
};

float viewportTransform[3][3];
cv::Rect2d viewportCrop;

class JSBridgeRenderer: public filmlab::ColorRenderer {
    public:
    float viewportTransform[3][3];

   Napi::Object getRendererData(Napi::Env env) {
      //printf("💚NODE: calculating matrices");
      Napi::Object jsData = Napi::Object::New(env);
      //calculateMatrices();
      
      Napi::Object lut = Napi::Object::New(env);      
      lut.Set("empty", workingColorSpaceToXYZLUT.data.empty());
      lut.Set("bins", workingColorSpaceToXYZLUT.bins);
      cv::Mat mat =  workingColorSpaceToXYZLUT.data;
      lut.Set("data", FloatArrayView(
         env,
         workingColorSpaceToXYZLUT.data.ptr<float>(),
         pow(workingColorSpaceToXYZLUT.bins, 3) * 3));
      lut.Set("inputMin", DoubleArrayView(env, workingColorSpaceToXYZLUT.inputMin, 3));
      lut.Set("inputMax", DoubleArrayView(env, workingColorSpaceToXYZLUT.inputMax, 3));

      jsData.Set("workingColorSpaceToXYZLUT", lut);

      jsData.Set("imageFormat", ImageFormatWrapper(env, formatData));

      //jsData.Set("inputTargetWhitePointXY", DoubleArrayView(env, inputTargetWhitePointXY, 2));
      jsData.Set("outputBrightness", outputBrightness);
      jsData.Set("outputContrast", outputContrast);
      jsData.Set("xyzToOutputColorTransformMatrix", DoubleArrayView(env, xyzToOutputColorTransformMatrix[0], 9));
      jsData.Set("workingColorSpaceLUTGammaCurve", static_cast<int>(workingColorSpaceLUTGammaCurve));

      jsData.Set("inputImageColorSpaceToWorkingColorSpaceMatrix", DoubleArrayView(env, inputImageColorSpaceToWorkingColorSpaceMatrix[0], 9));
            
      jsData.Set("imageFormatChanged", imageFormatChanged);
      jsData.Set("workingColorSpaceToXYZLUTChanged", workingColorSpaceToXYZLUTChanged);
      jsData.Set("adjustmentsChanged", adjustmentsChanged);

      // Calculate 3D transform from source image to viewport
      cv::Rect2d viewCrop;
      cv::Mat transform = sceneView->getSceneImageToViewportTransform(true, &viewCrop);
      cv::Size rendererSize = sceneView->getViewportSize();
      transform = filmlab::makeScaleMatrix(1.0 / rendererSize.width, 1.0 / rendererSize.height) *
          transform * filmlab::makeScaleMatrix(scene->inputImage.mat.cols, scene->inputImage.mat.rows);
      transform = filmlab::makeVerticalFlipMatrix(1.0) * transform; // Y-flipped for GL
      // Invert transform so it can be used as a lookup from output instead of a transform of input
      transform = filmlab::invertTransformMatrix(transform);
      transform.convertTo(transform, CV_32F);
      memcpy(viewportTransform, transform.ptr<float>(), 9 * sizeof(float));
      jsData.Set("viewTransform", FloatArrayView(env, viewportTransform[0], 9));
      jsData.Set("viewCropX", viewCrop.x);
      jsData.Set("viewCropY", viewCrop.y);
      jsData.Set("viewCropWidth", viewCrop.width);
      jsData.Set("viewCropHeight", viewCrop.height);
      jsData.Set("viewWidth", rendererSize.width);
      jsData.Set("viewHeight", rendererSize.height);

      return jsData;

   }
};

JSBridgeRenderer renderer;

void createScene() {
    if (scene != nullptr) {
        delete scene;
        scene = nullptr;
    }
    if (sceneView != nullptr) {
        delete sceneView;
        sceneView = nullptr;
    }
    scene = new filmlab::Scene();
    scene->useAsyncThreads = true;
    scene->useAutoMask = true;
    scene->autoSettingsCalculationInterval = 33;
    scene->changeDelegate = &sceneChangeHandler;
    sceneView = new filmlab::SceneView(scene);
    sceneView->changeDelegate = &rendererChangeHandler;
}

filmlab::AsyncImageLoader loaderX;
std::function<void(std::string, bool, filmlab::Image&)> loadHandler = [](const std::string path, bool success, filmlab::Image &image) {
  if (success) {
    printf("Successfully loaded image from %s\n", path.c_str());
    scene->setImage(image);
    currentImagePath = path;
    scene->applyChanges();
  }
};

void loadImage(const Napi::CallbackInfo& info) {
  if (loaderX.handler == nullptr) {
    loaderX.handler = &loadHandler;
  }
  std::string path = info[0].As<Napi::String>();
  printf("Queueing image load from from %s\n", path.c_str());
  // Abandon any previously queued images since we've already moved on.
  loaderX.clearQueue();
  loaderX.loadImageFromPath(path);
  scene->applyChanges();
}

Napi::Object getTextureImage(const Napi::CallbackInfo& info) {
  Napi::Env env = info.Env();
  return ImageWrapper(env, scene->inputImage);
}

Napi::String getCurrentImagePath(const Napi::CallbackInfo& info) {
  Napi::Env env = info.Env();
  return Napi::String::New(env, currentImagePath);
}

Napi::Boolean getImageLoaderIsBusy(const Napi::CallbackInfo& info) {
  Napi::Env env = info.Env();
  return Napi::Boolean::New(env, loaderX.isRunning());
}

Napi::String getCurrentImageOutputPath(const Napi::CallbackInfo& info) {
  Napi::Env env = info.Env();
  if (currentImagePath.length() == 0) return Napi::String::New(env, "");
  filmlab::ImageFileFormat outputFormat = (filmlab::ImageFileFormat)(info[0].As<Napi::Number>().Int32Value());
  std::string outputDir = info[1].As<Napi::String>();
  std::string newPath = filmlab::getSuggestedOutputFilename(currentImagePath, outputFormat, outputDir);
  return Napi::String::New(env, newPath);
}

class SaveImageWorker : public Napi::AsyncWorker {
public:
  Napi::Promise::Deferred deferred;
  SaveImageWorker(Napi::Env &env, std::string path, int bitDepth)
            : Napi::AsyncWorker(env), path(path), deferred(Napi::Promise::Deferred::New(env)), bitDepth(bitDepth) {
  }
  ~SaveImageWorker() {};

  void Execute() {
    if (currentImagePath.length() == 0) {
      SetError("No current image to save.");
    }
    cv::Mat out = scene->makeProcessedImage();
    bool success = filmlab::saveMat(out, path, bitDepth);
    if (!success) {
      SetError("Save failed.");
    }
    printf("🥶 WORKER: finished execute\n");
  }

  void OnOK() {
    printf("🥶 WORKER: ON OK\n");
    deferred.Resolve(Napi::Boolean::New(Env(), true));
  }

  void OnError(const Napi::Error& e) {
    printf("🥶 WORKER: ON Error\n");
    deferred.Resolve(Napi::Boolean::New(Env(), false));
  }

private:
  std::string path;
  int bitDepth;
};

Napi::Promise saveCurrentImage(const Napi::CallbackInfo& info) {
  Napi::Env env = info.Env();
  std::string path = info[0].As<Napi::String>();
  int bitDepth = info[1].As<Napi::Number>().Int32Value();

  SaveImageWorker* saver = new SaveImageWorker(env, path, bitDepth);
  saver->Queue();

  return saver->deferred.Promise();
}

void setViewportSize(const Napi::CallbackInfo& info) {
    int width = info[0].As<Napi::Number>().Int32Value();
    int height = info[1].As<Napi::Number>().Int32Value();
    if (sceneView != nullptr) {
        sceneView->setViewportSize(cv::Size(width, height));
        sceneView->setViewportSafeArea(cv::Rect2d(0, 0, width, height));
        sceneView->setViewportOrientation(filmlab::Orientation::AS_IS);
    }
    printf("Set view size %d, %d\n", width, height);
}

void setSceneSetting(const Napi::CallbackInfo& info) {
    int setting = info[0].As<Napi::Number>().Int32Value();
    double value = info[1].As<Napi::Number>().DoubleValue();
    scene->setSetting(filmlab::ProcessSetting(setting), value);
    scene->applyChanges();
}

napi_value getSceneSetting(const Napi::CallbackInfo& info) {
    Napi::Env env = info.Env();
    int setting = info[0].As<Napi::Number>().Int32Value();
    return Napi::Number::New(env, scene->getSetting(filmlab::ProcessSetting(setting)));
}

napi_value getImageWidth(const Napi::CallbackInfo& info) {
    Napi::Env env = info.Env();
    return Napi::Number::New(env, scene->inputImage.mat.cols);
}

napi_value getImageHeight(const Napi::CallbackInfo& info) {
    Napi::Env env = info.Env();
    return Napi::Number::New(env, scene->inputImage.mat.rows);
}

void zoomIn(const Napi::CallbackInfo& info) {
  sceneView->adjustViewportZoom(1.5);
}

void zoomOut(const Napi::CallbackInfo& info) {
  sceneView->adjustViewportZoom(1.0/1.5);
}

void setZoom(const Napi::CallbackInfo& info) {
  double zoom = info[0].As<Napi::Number>().DoubleValue();
  sceneView->adjustViewportZoom(zoom);
}

void adjustPan(const Napi::CallbackInfo& info) {
  assert(info.Length() == 2);
  double x = info[0].As<Napi::Number>().DoubleValue();
  double y = info[1].As<Napi::Number>().DoubleValue();
  sceneView->adjustViewportPan(cv::Point2d(x, y));
}

void flipHorizontally(const Napi::CallbackInfo& info) {
  scene->settings.flipHorizontally();
  scene->applyChanges();
}

void rotateCounterClockwise(const Napi::CallbackInfo& info) {
  scene->settings.rotateCounterClockwise();
  scene->applyChanges();
}

napi_value hasColorControl(const Napi::CallbackInfo& info) {
    if (scene == nullptr || scene->process == nullptr) {
        return Napi::Boolean::New(info.Env(), false);
    }
    return Napi::Boolean::New(info.Env(),
        scene->process->hasColorTempControl() &&
        scene->process->hasColorTintControl());
}

napi_value hasExposureControl(const Napi::CallbackInfo& info) {
    if (scene == nullptr || scene->process == nullptr) {
        return Napi::Boolean::New(info.Env(), false);
    }
    return Napi::Boolean::New(info.Env(),
        scene->process->hasExposureControl());
}

Napi::Object getRendererData(const Napi::CallbackInfo& info) {
    if (scene != nullptr && scene->process != nullptr) {
        scene->process->updateRenderer(scene->inputImage.format, &renderer);
    }
    return renderer.getRendererData(info.Env());
}

Napi::Boolean getSceneDataChanged(const Napi::CallbackInfo& info) {
  Napi::Boolean result = Napi::Boolean::New(info.Env(), sceneDataChanged);
  sceneDataChanged = false;
  return result;
}

Napi::Boolean getRendererDataChanged(const Napi::CallbackInfo& info) {
  Napi::Boolean result = Napi::Boolean::New(info.Env(), rendererDataChanged);
  rendererDataChanged = false;
  return result;
}

Napi::Object Init(Napi::Env env, Napi::Object exports) {
  printf("FilmLab native layer initializing.\n");

  createScene();

  exports.Set("loadImage",
              Napi::Function::New(env, loadImage));
              
  exports.Set("getImageLoaderIsBusy",
              Napi::Function::New(env, getImageLoaderIsBusy));

  exports.Set("getCurrentImagePath",
              Napi::Function::New(env, getCurrentImagePath));

  exports.Set("getCurrentImageOutputPath",
              Napi::Function::New(env, getCurrentImageOutputPath));

  exports.Set("saveCurrentImage",
              Napi::Function::New(env, saveCurrentImage));

  exports.Set("setViewportSize",
              Napi::Function::New(env, setViewportSize));

  exports.Set("setSceneSetting",
              Napi::Function::New(env, setSceneSetting));

  exports.Set("getSceneSetting",
              Napi::Function::New(env, getSceneSetting));

  exports.Set("getImageWidth",
              Napi::Function::New(env, getImageWidth));

  exports.Set("getImageHeight",
              Napi::Function::New(env, getImageHeight));

  exports.Set("setZoom",
              Napi::Function::New(env, setZoom));

  exports.Set("zoomIn",
              Napi::Function::New(env, zoomIn));

  exports.Set("zoomOut",
              Napi::Function::New(env, zoomOut));  

  exports.Set("adjustPan",
              Napi::Function::New(env, adjustPan));

  exports.Set("flipHorizontally",
              Napi::Function::New(env, flipHorizontally));

  exports.Set("rotateCounterClockwise",
              Napi::Function::New(env, rotateCounterClockwise));  

  exports.Set("hasColorControl",
              Napi::Function::New(env, hasColorControl));

  exports.Set("hasExposureControl",
              Napi::Function::New(env, hasExposureControl));  

  exports.Set("getRendererData", Napi::Function::New(env, getRendererData));

  exports.Set("getTextureImage", Napi::Function::New(env, getTextureImage));  

  exports.Set("sceneDataChanged", Napi::Function::New(env, getSceneDataChanged));
  exports.Set("rendererDataChanged", Napi::Function::New(env, getRendererDataChanged));

  return exports;
}

NODE_API_MODULE(NODE_GYP_MODULE_NAME, Init)
